@@ -3,6 +3,7 @@ use crate::{
     system_instruction_processor,
 };
 use solana_sdk::{
+    feature_set,
     instruction::InstructionError,
     process_instruction::{stable_log, InvokeContext, ProcessInstructionWithContext},
     pubkey::Pubkey,
@@ -86,7 +87,15 @@ pub enum ActivationType {
 /// normal child Bank creation.
 /// https://github.com/solana-labs/solana/blob/84b139cc94b5be7c9e0c18c2ad91743231b85a0d/runtime/src/bank.rs#L1723
 fn feature_builtins() -> Vec<(Builtin, Pubkey, ActivationType)> {
-    vec![]
+    vec![(
+        Builtin::new(
+            "compute_budget_program",
+            solana_sdk::compute_budget::id(),
+            solana_compute_budget_program::process_instruction,
+        ),
+        feature_set::tx_wide_compute_cap::id(),
+        ActivationType::NewProgram,
+    )]
 }
 
 pub(crate) fn get() -> Builtins {
